package at.bestsolution.rsd.rt;

import java.util.ArrayList;
import java.util.List;
import java.util.function.BiPredicate;
import java.util.function.Function;
import java.util.function.ToIntFunction;
import java.util.stream.Collector;
import java.util.stream.Stream;

public interface Array<T> extends Iterable<T> {
	public interface DoubleArray extends Array<Double> {
		double getAsDouble(int index);
		double[] toDoubleArray();
		int deepHashCode();

		public static int computeDeepHashCode(DoubleArray obj) {
			if( obj == null ) {
				return 0;
			}
			return obj.deepHashCode();
		}

		boolean deepEquals(DoubleArray array);
		public static boolean computeDeepEquals(DoubleArray a, DoubleArray b) {
			if( a == b ) {
				return true;
			} else if( a == null ) {
				return false;
			}
			return a.deepEquals(b);
		}

		public static DoubleArray of(double... array) {
			if( array.length == 0 ) {
				return emptyDouble();
			}
			return new DoubleArrayImpl(array);
		}

	}

	public interface LongArray extends Array<Long> {
		long getAsLong(int index);
		long[] toLongArray();
		int deepHashCode();

		public static int computeDeepHashCode(LongArray obj) {
			if( obj == null ) {
				return 0;
			}
			return obj.deepHashCode();
		}

		boolean deepEquals(LongArray array);
		public static boolean computeDeepEquals(LongArray a, LongArray b) {
			if( a == b ) {
				return true;
			} else if( a == null ) {
				return false;
			}
			return a.deepEquals(b);
		}

		public static LongArray of(long... array) {
			if( array.length == 0 ) {
				return emptyLong();
			}
			return new LongArrayImpl(array);
		}
	}

	public interface IntArray extends Array<Integer> {
		int getAsInt(int index);
		int[] toIntArray();
		int deepHashCode();

		public static int computeDeepHashCode(LongArray obj) {
			if( obj == null ) {
				return 0;
			}
			return obj.deepHashCode();
		}

		boolean deepEquals(IntArray array);
		public static boolean computeDeepEquals(IntArray a, IntArray b) {
			if( a == b ) {
				return true;
			} else if( a == null ) {
				return false;
			}
			return a.deepEquals(b);
		}

		public static IntArray of(int... array) {
			if( array.length == 0 ) {
				return emptyInt();
			}
			return new IntArrayImpl(array);
		}
	}

	public int length();

	public boolean isEmpty();

	public <S> S[] toArray(S[] t);

	public T get(int index);

	public Stream<T> stream();

	public int deepHashCode(ToIntFunction<T> function);

	public static <T> int computeDeepHashCode(Array<T> obj, ToIntFunction<T> function) {
		if( obj == null ) {
			return 0;
		}
		return obj.deepHashCode(function);
	}

	public boolean deepEquals(Array<T> array, BiPredicate<T, T> function);
	public static <T> boolean computeDeepEquals(Array<T> a, Array<T> b, BiPredicate<T, T> function) {
		if( a == b ) {
			return true;
		} else if( a == null ) {
			return false;
		}
		return a.deepEquals(b, function);
	}

	public default String toString(Function<T, String> function) {
		StringBuilder b = new StringBuilder("[\n");
		for( int i = 0; i < length(); i++ ) {
			b.append("    " + function.apply(get(i)));
			if( i + 1 < length() ) {
				b.append(", \n");
			}
		}
		b.append("\n]");
		return b.toString();
	}

	/**
	 * Map the array to an unmodifiable list
	 *
	 * @param <R> the target type
	 * @param map function to map to the new type
	 * @return the list
	 */
	public <R> List<R> mapTo(Function<T, R> map);

	public default List<T> toList() {
		List<T> rv = new ArrayList<>(length());
		for( int i = 0; i < length(); i++ ) {
			rv.add(get(i));
		}
		return rv;
	}

	@SuppressWarnings("unchecked")
	public static <T> Array<T> empty() {
		return (Array<T>) EmptyArray.SELF;
	}

	public static Array.DoubleArray emptyDouble() {
		return EmptyDoubleArray.SELF;
	}

	public static Array.IntArray emptyInt() {
		return EmptyIntArray.SELF;
	}

	public static Array.LongArray emptyLong() {
		return EmptyLongArray.SELF;
	}

	public interface Builder<T> {
		public Builder<T> add(int index, T t);

		public Builder<T> addMany(int index, List<T> t);

		public Builder<T> remove(T t);

		public Builder<T> removeMany(List<T> t);

		public Array<T> build();
	}

	@Deprecated
	public static <T> Array<T> from(List<T> array) {
		return of(array);
	}

	@Deprecated
	public static <T> Array<T> from(T[] array) {
		return of(array);
	}

	@Deprecated
	public static DoubleArray from(double[] array) {
		return DoubleArray.of(array);
	}

	public static <T> Collector<T, ?, Array<T>> toArray() {
		return new ArrayCollector<>();
	}

	@SuppressWarnings("unchecked")
	public static <T> Array<T> of(List<T> array) {
		return (Array<T>) of(array.toArray());
	}

	@SuppressWarnings("unchecked")
	public static <T> Array<T> of(T... array) {
		if (array.length == 0) {
			return empty();
		}

		return new ArrayImpl<>(array, true);
	}
}
