package at.bestsolution.rsd.rt;

import java.io.Serializable;
import java.util.Objects;
import java.util.OptionalInt;

public class RSDException extends RuntimeException {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	
	private final int statusCode;
	
	private final long errorCode;
	
	private final CharSequence errorText;
	
	static <T> T assertNoneAndSerializable(T value) {
		Objects.requireNonNull(value);
		if( ! (value instanceof Serializable) ) {
			throw new IllegalArgumentException( value.getClass() + " is not serializable.");
		}
		return value;
	}
	
	public RSDException(long errorCode, CharSequence errorText) {
		super(Objects.toString(errorText));
		this.errorText = assertNoneAndSerializable(errorText);
		this.errorCode = errorCode;
		this.statusCode = -1;
	}
		
	public RSDException(Throwable cause, long errorCode, CharSequence errorText) {
		super(Objects.toString(errorText), cause);
		this.errorText = assertNoneAndSerializable(errorText);
		this.errorCode = errorCode;
		this.statusCode = -1;
	}
	
	public RSDException(String message, long errorCode, CharSequence errorText) {
		super(message);
		this.errorText = assertNoneAndSerializable(errorText);
		this.errorCode = errorCode;
		this.statusCode = -1;
	}
	
	public RSDException(String message, Throwable cause, long errorCode, CharSequence errorText) {
		super(message, cause);
		this.errorText = assertNoneAndSerializable(errorText);
		this.errorCode = errorCode;
		this.statusCode = -1;
	}
	
	public RSDException(String message, Throwable cause, int statusCode, long errorCode, CharSequence errorText) {
		super(message, cause);
		this.errorText = assertNoneAndSerializable(errorText);
		this.errorCode = errorCode;
		this.statusCode = statusCode;
	}
	
	public RSDException(ErrorType<?> errorType, String message, Throwable cause, CharSequence errorText) {
		super(message, cause);
		this.errorText = assertNoneAndSerializable(errorText);
		this.errorCode = errorType.errorCode;
		this.statusCode = errorType.statusCode;
	}
	
	public long errorCode() {
		return errorCode;
	}
	
	public CharSequence errorText() {
		return errorText;
	}
	
	public OptionalInt statusCode() {
		if( statusCode != -1 ) {
			return OptionalInt.of(statusCode);
		}
		return OptionalInt.empty();
	}
	
	public RSDErrorResult toResultObject() {
		return new RSDErrorResult(statusCode, errorCode, Objects.toString(errorText));
	}
	
	public abstract static class Builder<T extends RSDException> {
		protected int statusCode;
		
		protected long errorCode;
		
		protected CharSequence errorText;
		
		protected String message;
		
		protected Throwable cause;
		
		public Builder<T> statusCode(int statusCode) {
			this.statusCode = statusCode;
			return this;
		}
		
		public Builder<T> errorCode(long errorCode) {
			this.errorCode = errorCode;
			return this;
		}
		
		public Builder<T> errorText(CharSequence errorText) {
			this.errorText = errorText;
			return this;
		}
		
		public Builder<T> message(String message) {
			this.message = message;
			return this;
		}
		
		public Builder<T> cause(Throwable cause) {
			this.cause = cause;
			return this;
		}
		
		public abstract T build();
	}
}
