package at.bestsolution.rsd.rt.poll;

import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Consumer;
import java.util.function.Function;

import at.bestsolution.rsd.rt.Subscription;

class FastPollingRegistry implements PollingRegistry {
	private ScheduledExecutorService executorService; //lazily initialised when subscriptions are added
	private final AtomicLong activeSubscriptions = new AtomicLong(0);
	private static long POLLING_TIMEOUT;

	static {
		try {
			POLLING_TIMEOUT = Long.valueOf(System.getProperty("rsd.fastpolling.frequency", "1000"));
		} catch (Throwable t) {
			POLLING_TIMEOUT = 1000;
		}
	}

	private final long delay;

	public FastPollingRegistry() {
		this.delay = POLLING_TIMEOUT;
	}

	public FastPollingRegistry(long frequence) {
		this.delay = frequence;
	}

	private synchronized ScheduledExecutorService getOrCreateExecutor() {
		if (!isExecutorRunning()) {
			executorService = Executors.newScheduledThreadPool(4, code -> {
				Thread t = new Thread(code);
				t.setDaemon(true);
				//set name, useful when debugging
				t.setName(this.getClass().getSimpleName() + "-" + t.getName());
				return t;
			});
		}
		return executorService;
	}

	@Override
	public <T> Subscription register(Function<String, PollResult<T>> remoteCall, Consumer<T> consumer) {
		getOrCreateExecutor();
		AtomicReference<String> stateMarker = new AtomicReference<>();
		ScheduledFuture<?> future = getOrCreateExecutor().scheduleWithFixedDelay(() -> {
			PollResult<T> result = remoteCall.apply(stateMarker.get());
			stateMarker.set(result.stateMarker());
			result.result().forEach(consumer::accept);
		}, 0, delay, TimeUnit.MILLISECONDS);
		return new SelfRemovingSubscription(future);
	}

	@Override
	public Subscription registerVoid(Function<String, PollResult<Void>> remoteCall, Runnable consumer) {
		AtomicReference<String> stateMarker = new AtomicReference<>();
		ScheduledFuture<?> future = getOrCreateExecutor().scheduleWithFixedDelay(() -> {
			PollResult<Void> result = remoteCall.apply(stateMarker.get());
			stateMarker.set(result.stateMarker());
			consumer.run();
		}, 0, delay, TimeUnit.MILLISECONDS);
		return new SelfRemovingSubscription(future);
	}
	
	boolean isExecutorRunning() {
		return executorService != null;
	}

	class SelfRemovingSubscription implements Subscription {
		private final ScheduledFuture<?> future;

		public SelfRemovingSubscription(ScheduledFuture<?> future) {
			this.future = future;
			activeSubscriptions.incrementAndGet();
		}

		@Override
		public synchronized void dispose() {
			if (!future.isCancelled()) {
				future.cancel(false);
				if (activeSubscriptions.decrementAndGet() == 0) {
					executorService.shutdown();
					executorService = null;
				}
			}
		}
	}

}
