package at.bestsolution.rsd.rt.rest;

import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import java.util.Optional;
import java.util.concurrent.CompletionException;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Function;

import at.bestsolution.rsd.rt.InterceptorRegistry;
import at.bestsolution.rsd.rt.interceptor.RemoteCallInterceptor.CallContext;
import at.bestsolution.rsd.rt.interceptor.RemoteResultInterceptor.ResultContext;

class RemoteMethodInvocationImpl<T> implements RemoteMethodInvocation<T> {
	private final Request request;
	private final InterceptorRegistry registry;
	private final String methodName;

	public RemoteMethodInvocationImpl(String methodName, Request request, InterceptorRegistry registry) {
		this.request = request;
		this.registry = registry;
		this.methodName = methodName;
	}

	public T invoke(Function<ResponseStream, T> stream) {
		preInvoke(methodName);
		try {
			ResponseStream r = request.execute();
			T value = stream.apply(r);
			return postInvoke(methodName, value, null);
		} catch (RuntimeException e) {
			return postInvoke(methodName, null, e);
		}
	}

	private T postInvoke(String methodName, T value, RuntimeException ex) {
		AtomicBoolean recovered = new AtomicBoolean(ex == null);
		ResultContext context = new ResultContext() {

			@Override
			public Optional<Object> getProperty(String key) {
				return Optional.empty();
			}

			@Override
			public String getMethod() {
				return methodName;
			}

			@Override
			public void recovered() {
				recovered.set(true);
			}
		};
		T result = registry.processResult(value, ex, context);
		if( ex != null && ! recovered.get() ) {
			if( ex instanceof CompletionException && ex.getCause() instanceof RuntimeException ) {
				throw (RuntimeException)ex.getCause();
			}
			throw ex;
		}
		return result;
	}

	private void preInvoke(String methodName) {
		CallContext context = new CallContext() {

			@SuppressWarnings("unchecked")
			@Override
			public void setProperty(String key, Object o) {
				if (o instanceof Map<?, ?>) {
					StringBuilder b = new StringBuilder();
					b.append('{');
					boolean flag = false;
					for( Entry<Object, Object> e : ((Map<Object,Object>)o).entrySet() ) {
						if( flag ) {
							b.append(", ");
						}
						b.append( escape(Objects.toString(e.getKey())));
						b.append(" : ");
						b.append( escape(Objects.toString(e.getValue())));
						flag = true;
					}
					b.append('}');
					request.setHeader(key, b.toString());
				} else {
					request.setHeader(key, Objects.toString(o));
				}
			}

			@Override
			public String getMethod() {
				return methodName;
			}

			@Override
			public String getComponentType() {
				return "rest";
			}
		};
		registry.forEachCallInterceptor(i -> i.preInvokation(context));
	}

	static char[] escape(CharSequence val) {
		StringBuilder b = new StringBuilder(val.length() + 2);
		b.append('"');

		char c;
		for (int i = 0; i < val.length(); i++) {
			c = val.charAt(i);
			switch (c) {
			case '\\':
			case '"':
				b.append('\\');
				b.append(c);
				break;
			case '\b':
				b.append("\\b");
				break;
			case '\t':
				b.append("\\t");
				break;
			case '\n':
				b.append("\\n");
				break;
			case '\f':
				b.append("\\f");
				break;
			case '\r':
				b.append("\\r");
				break;
			default:
				if (c < ' ' || (c >= '\u0080' && c < '\u00a0') || (c >= '\u2000' && c < '\u2100')) {
					b.append("\\u");
					String hex = Integer.toHexString(c);
					b.append("0000", 0, 4 - hex.length());
					b.append(hex);
				} else {
					b.append(c);
				}
			}
		}
		b.append('"');

		char[] rv = new char[b.length()];
		b.getChars(0, b.length(), rv, 0);
		return rv;
	}
}
