/*******************************************************************************
 * Copyright (c) 2003, 2017 IBM Corporation and others.
 *
 * This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *     Tom Schindl<tom.schindl@bestsolution.at> - adaption to APIs
 *******************************************************************************/

package at.bestsolution.rsd.rt.util.internal;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import at.bestsolution.rsd.rt.util.LdapFilterSupport.BranchNode;
import at.bestsolution.rsd.rt.util.LdapFilterSupport.FilterNode;
import at.bestsolution.rsd.rt.util.LdapFilterSupport.LeafNode;
import at.bestsolution.rsd.rt.util.LdapFilterSupport.LeafNodeSubstring;
import at.bestsolution.rsd.rt.util.LdapFilterSupport.NodeType;

// Taken from org.eclipse.osgi.internal.framework.FilterImpl

public final class Parser {
	private final String filterstring;
	private final char[] filterChars;
	private int pos;

	public Parser(String filterstring) {
		this.filterstring = filterstring;
		filterChars = filterstring.toCharArray();
		pos = 0;
	}

	public FilterNode parse() throws InvalidSyntaxException {
		FilterNode filter;
		try {
			filter = parse_filter();
		} catch (ArrayIndexOutOfBoundsException e) {
			throw new InvalidSyntaxException(Msg.FILTER_TERMINATED_ABRUBTLY, filterstring, e);
		}

		if (pos != filterChars.length) {
			throw new InvalidSyntaxException(String.format(Msg.FILTER_TRAILING_CHARACTERS, filterstring.substring(pos)),
					filterstring);
		}
		return filter;
	}

	private FilterNode parse_filter() throws InvalidSyntaxException {
		FilterNode filter;
		skipWhiteSpace();

		if (filterChars[pos] != '(') {
			throw new InvalidSyntaxException(String.format(Msg.FILTER_MISSING_LEFTPAREN, filterstring.substring(pos)),
					filterstring);
		}

		pos++;

		filter = parse_filtercomp();

		skipWhiteSpace();

		if (filterChars[pos] != ')') {
			throw new InvalidSyntaxException(String.format(Msg.FILTER_MISSING_RIGHTPAREN, filterstring.substring(pos)),
					filterstring);
		}

		pos++;

		skipWhiteSpace();

		return filter;
	}

	private FilterNode parse_filtercomp() throws InvalidSyntaxException {
		skipWhiteSpace();

		char c = filterChars[pos];

		switch (c) {
		case '&': {
			pos++;
			return parse_and();
		}
		case '|': {
			pos++;
			return parse_or();
		}
		case '!': {
			pos++;
			return parse_not();
		}
		}
		return parse_item();
	}

	private FilterNode parse_and() throws InvalidSyntaxException {
		int lookahead = pos;
		skipWhiteSpace();

		if (filterChars[pos] != '(') {
			pos = lookahead - 1;
			return parse_item();
		}

		List<FilterNode> operands = new ArrayList<>(10);

		while (filterChars[pos] == '(') {
			FilterNode child = parse_filter();
			operands.add(child);
		}

		return new BranchNodeImpl(NodeType.AND, null, operands.toArray(new FilterNode[0]));
	}

	private FilterNode parse_or() throws InvalidSyntaxException {
		int lookahead = pos;
		skipWhiteSpace();

		if (filterChars[pos] != '(') {
			pos = lookahead - 1;
			return parse_item();
		}

		List<FilterNode> operands = new ArrayList<>(10);

		while (filterChars[pos] == '(') {
			FilterNode child = parse_filter();
			operands.add(child);
		}

		return new BranchNodeImpl(NodeType.OR, null, operands.toArray(new FilterNode[0]));
	}

	private FilterNode parse_not() throws InvalidSyntaxException {
		int lookahead = pos;
		skipWhiteSpace();

		if (filterChars[pos] != '(') {
			pos = lookahead - 1;
			return parse_item();
		}

		FilterNode child = parse_filter();

		return new BranchNodeImpl(NodeType.NOT, null, new FilterNode[] { child });
	}

	private FilterNode parse_item() throws InvalidSyntaxException {
		String attr = parse_attr();

		skipWhiteSpace();

		switch (filterChars[pos]) {
		case '~': {
			if (filterChars[pos + 1] == '=') {
				pos += 2;
				return new LeafNodeImpl(NodeType.APPROX, attr, parse_value());
			}
			break;
		}
		case '>': {
			if (filterChars[pos + 1] == '=') {
				pos += 2;
				return new LeafNodeImpl(NodeType.GREATER, attr, parse_value());
			}
			break;
		}
		case '<': {
			if (filterChars[pos + 1] == '=') {
				pos += 2;
				return new LeafNodeImpl(NodeType.LESS, attr, parse_value());
			}
			break;
		}
		case '=': {
			if (filterChars[pos + 1] == '*') {
				int oldpos = pos;
				pos += 2;
				skipWhiteSpace();
				if (filterChars[pos] == ')') {
					return new LeafNodeImpl(NodeType.PRESENT, attr, null);
				}
				pos = oldpos;
			}

			pos++;
			Object string = parse_substring();

			if (string instanceof String) {
				return new LeafNodeImpl(NodeType.EQUAL, attr, (String) string);
			}
			return new LeafNodeSubstringImpl(attr, (String[]) string);
		}
		}

		throw new InvalidSyntaxException(String.format(Msg.FILTER_INVALID_OPERATOR, filterstring.substring(pos)),
				filterstring);
	}

	private String parse_attr() throws InvalidSyntaxException {
		skipWhiteSpace();

		int begin = pos;
		int end = pos;

		char c = filterChars[pos];

		while (c != '~' && c != '<' && c != '>' && c != '=' && c != '(' && c != ')') {
			pos++;

			if (!Character.isWhitespace(c)) {
				end = pos;
			}

			c = filterChars[pos];
		}

		int length = end - begin;

		if (length == 0) {
			throw new InvalidSyntaxException(String.format(Msg.FILTER_MISSING_ATTR, filterstring.substring(pos)),
					filterstring);
		}

		return new String(filterChars, begin, length);
	}

	private String parse_value() throws InvalidSyntaxException {
		StringBuilder sb = new StringBuilder(filterChars.length - pos);

		parseloop: while (true) {
			char c = filterChars[pos];

			switch (c) {
			case ')': {
				break parseloop;
			}

			case '(': {
				throw new InvalidSyntaxException(String.format(Msg.FILTER_INVALID_VALUE, filterstring.substring(pos)),
						filterstring);
			}

			case '\\': {
				pos++;
				c = filterChars[pos];
				/* fall through into default */
			}

			default: {
				sb.append(c);
				pos++;
				break;
			}
			}
		}

		if (sb.length() == 0) {
			throw new InvalidSyntaxException(String.format(Msg.FILTER_MISSING_VALUE, filterstring.substring(pos)),
					filterstring);
		}

		return sb.toString();
	}

	private Object parse_substring() throws InvalidSyntaxException {
		StringBuilder sb = new StringBuilder(filterChars.length - pos);

		List<String> operands = new ArrayList<>(10);

		parseloop: while (true) {
			char c = filterChars[pos];

			switch (c) {
			case ')': {
				if (sb.length() > 0) {
					operands.add(sb.toString());
				}

				break parseloop;
			}

			case '(': {
				throw new InvalidSyntaxException(String.format(Msg.FILTER_INVALID_VALUE, filterstring.substring(pos)),
						filterstring);
			}

			case '*': {
				if (sb.length() > 0) {
					operands.add(sb.toString());
				}

				sb.setLength(0);

				operands.add(null);
				pos++;

				break;
			}

			case '\\': {
				pos++;
				c = filterChars[pos];
				/* fall through into default */
			}

			default: {
				sb.append(c);
				pos++;
				break;
			}
			}
		}

		int size = operands.size();

		if (size == 0) {
			return ""; //$NON-NLS-1$
		}

		if (size == 1) {
			Object single = operands.get(0);

			if (single != null) {
				return single;
			}
		}

		return operands.toArray(new String[size]);
	}

	private void skipWhiteSpace() {
		for (int length = filterChars.length; (pos < length) && Character.isWhitespace(filterChars[pos]);) {
			pos++;
		}
	}

	abstract static class FilterImpl<T> implements FilterNode {
		final NodeType op;
		/** filter attribute or null if operation AND, OR or NOT */
		final String attr;
		/** filter operands */
		final T value;

		private transient volatile String filterString;

		public FilterImpl(NodeType op, String attr, T value) {
			this.op = op;
			this.attr = attr;
			this.value = value;
		}

		/**
		 * Returns this <code>Filter</code> object's filter string.
		 * <p>
		 * The filter string is normalized by removing whitespace which does not affect
		 * the meaning of the filter.
		 *
		 * @return Filter string.
		 */
		@Override
		public String toString() {
			String result = filterString;
			if (result == null) {
				filterString = result = normalize().toString();
			}
			return result;
		}

		/**
		 * Encode the value string such that '(', '*', ')' and '\' are escaped.
		 *
		 * @param value unencoded value string.
		 * @return encoded value string.
		 */
		static String encodeValue(String value) {
			boolean encoded = false;
			int inlen = value.length();
			int outlen = inlen << 1; /* inlen * 2 */

			char[] output = new char[outlen];
			value.getChars(0, inlen, output, inlen);

			int cursor = 0;
			for (int i = inlen; i < outlen; i++) {
				char c = output[i];

				switch (c) {
				case '(':
				case '*':
				case ')':
				case '\\': {
					output[cursor] = '\\';
					cursor++;
					encoded = true;

					break;
				}
				}

				output[cursor] = c;
				cursor++;
			}

			return encoded ? new String(output, 0, cursor) : value;
		}

		/**
		 * Returns this <code>Filter</code>'s normalized filter string.
		 * <p>
		 * The filter string is normalized by removing whitespace which does not affect
		 * the meaning of the filter.
		 * 
		 * @return This <code>Filter</code>'s filter string.
		 */
		private StringBuilder normalize() {
			StringBuilder sb = new StringBuilder();
			sb.append('(');

			switch (op) {
			case AND: {
				sb.append('&');

				FilterImpl[] filters = (FilterImpl[]) value;
				for (FilterImpl f : filters) {
					sb.append(f.normalize());
				}

				break;
			}

			case OR: {
				sb.append('|');

				FilterImpl[] filters = (FilterImpl[]) value;
				for (FilterImpl f : filters) {
					sb.append(f.normalize());
				}

				break;
			}

			case NOT: {
				sb.append('!');
				FilterImpl[] filter = (FilterImpl[]) value;
				sb.append(filter[0].normalize());

				break;
			}

			case SUBSTRING: {
				sb.append(attr);
				sb.append('=');

				String[] substrings = (String[]) value;

				for (String substr : substrings) {
					if (substr == null) /* * */ {
						sb.append('*');
					} else /* xxx */ {
						sb.append(encodeValue(substr));
					}
				}

				break;
			}
			case EQUAL: {
				sb.append(attr);
				sb.append('=');
				sb.append(encodeValue((String) value));

				break;
			}
			case GREATER: {
				sb.append(attr);
				sb.append(">="); //$NON-NLS-1$
				sb.append(encodeValue((String) value));

				break;
			}
			case LESS: {
				sb.append(attr);
				sb.append("<="); //$NON-NLS-1$
				sb.append(encodeValue((String) value));

				break;
			}
			case APPROX: {
				sb.append(attr);
				sb.append("~="); //$NON-NLS-1$
				sb.append(encodeValue(approxString((String) value)));

				break;
			}

			case PRESENT: {
				sb.append(attr);
				sb.append("=*"); //$NON-NLS-1$

				break;
			}
			}

			sb.append(')');

			return sb;
		}

		private static String approxString(String input) {
			boolean changed = false;
			char[] output = input.toCharArray();
			int cursor = 0;
			for (char c : output) {
				if (Character.isWhitespace(c)) {
					changed = true;
					continue;
				}

				output[cursor] = c;
				cursor++;
			}

			return changed ? new String(output, 0, cursor) : input;
		}

		@Override
		public NodeType type() {
			return op;
		}
	}

	static final class BranchNodeImpl extends FilterImpl<FilterNode[]> implements BranchNode {

		public BranchNodeImpl(NodeType op, String attr, FilterNode[] value) {
			super(op, attr, value);
		}

		@Override
		public List<FilterNode> children() {
			return Arrays.asList((FilterNode[]) this.value);
		}
	}

	static final class LeafNodeImpl extends FilterImpl<String> implements LeafNode {

		public LeafNodeImpl(NodeType op, String attr, String value) {
			super(op, attr, value);
		}

		@Override
		public String value() {
			return this.value;
		}

		@Override
		public String attribute() {
			return this.attr;
		}
	}

	static class LeafNodeSubstringImpl extends FilterImpl<String[]> implements LeafNodeSubstring {
		private transient volatile String valueString;

		public LeafNodeSubstringImpl(String attr, String[] value) {
			super(NodeType.SUBSTRING, attr, value);
		}

		@Override
		public String value() {
			if (valueString != null) {
				return valueString;
			}

			StringBuilder sb = new StringBuilder();
			String[] substrings = (String[]) value;

			for (String substr : substrings) {
				if (substr == null) /* * */ {
					sb.append('*');
				} else /* xxx */ {
					sb.append(encodeValue(substr));
				}
			}

			return valueString = sb.toString();
		}

		@Override
		public List<String> segments() {
			return Arrays.asList((String[]) value);
		}

		@Override
		public String attribute() {
			return this.attr;
		}
	}
}